#include <iostream>
#include <CompLib.pb.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <zconf.h>
#include "include/messageBuilder.hpp"
#include "include/errorMessages.hpp"
#include <spdlog/spdlog.h>
#include <chrono>

#include "include/spi.hpp"
#include "include/reset.hpp"
#include "include/encoder.hpp"
#include "include/motor.hpp"
#include "include/robot.hpp"

#define SOCKET_PATH "/tmp/compLib"
#define BUFFER_SIZE 64

google::protobuf::Message *processMessage(const std::string &serializedMessage) {
  CompLib::GenericRequest message;
  message.ParseFromString(serializedMessage);
  auto messageTypeName = message.header().message_type();

  if (messageTypeName == CompLib::ReadSensorsRequest::GetDescriptor()->full_name()) {
    CompLib::ReadSensorsRequest readSensorsRequest;
    readSensorsRequest.ParseFromString(serializedMessage);
  } else if (messageTypeName == CompLib::ReadSensorsResponse::GetDescriptor()->full_name()) {
    CompLib::ReadSensorsResponse readSensorsResponse;
    readSensorsResponse.ParseFromString(serializedMessage);
    std::cout << readSensorsResponse.ir_1() << std::endl;
    std::cout << readSensorsResponse.ir_2() << std::endl;
    std::cout << readSensorsResponse.ir_3() << std::endl;
    std::cout << readSensorsResponse.ir_4() << std::endl;
    std::cout << readSensorsResponse.ir_5() << std::endl;
  } else {
    std::cout << messageTypeName << " not found!" << std::endl;
  }

  google::protobuf::Message *returnMessage = MessageBuilder::genericResponse(false, ERROR_MESSAGE_UNKNOWN);
  return returnMessage;
}

[[noreturn]] void socketServer() {
  struct sockaddr_un socketAddress;
  int socketFileDescriptor = socket(AF_UNIX, SOCK_STREAM, 0);
  remove(SOCKET_PATH);

  memset(&socketAddress, 0, sizeof(struct sockaddr_un));
  socketAddress.sun_family = AF_UNIX;
  strncpy(socketAddress.sun_path, SOCKET_PATH, sizeof(socketAddress.sun_path) - 1);

  if (bind(socketFileDescriptor, (struct sockaddr *) &socketAddress, sizeof(struct sockaddr_un)) == -1) {
    exit(-1);
  }

  if (listen(socketFileDescriptor, 1) == -1) {
    exit(-2);
  }

  char readBuffer[BUFFER_SIZE];
  char writeBuffer[BUFFER_SIZE];
  for (;;) {
    int clientFileDescriptor = accept(socketFileDescriptor, NULL, NULL);
    auto numRead = read(clientFileDescriptor, readBuffer, 1);
    std::cout << numRead << std::endl;

    uint8_t messageSize = readBuffer[0];
    std::cout << std::to_string(messageSize) << std::endl;

    numRead = read(clientFileDescriptor, readBuffer, readBuffer[0]);
    std::cout << numRead << std::endl;

    std::string stringMessage;
    for (int i{}; i < messageSize; i++) {
      stringMessage += readBuffer[i];
    }

    auto response = processMessage(stringMessage);
    uint8_t responseSize = response->ByteSizeLong();
    writeBuffer[0] = responseSize;

    std::cout << std::to_string(responseSize) << std::endl;
    write(clientFileDescriptor, writeBuffer, 1);

    response->SerializeToArray(writeBuffer, BUFFER_SIZE);
    write(clientFileDescriptor, writeBuffer, responseSize);

    close(clientFileDescriptor);
  }

}

int main() {
  // socketServer();
  Reset::reset_robot();
  // Spi &spi = Spi::getInstance();
  // int version = spi.read(1, 1);

  // double count = 20000;
  // double start_time = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::system_clock::now().time_since_epoch()).count();
  // for (int i{0}; i < count; i++){
  //     spi.read(1, 1);
  // }
  // double end_time = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::system_clock::now().time_since_epoch()).count();

  double power = 0;
  // int port = 0;
  // long double accumulator = 0;
  // long double alpha = 0.2;
  while (power < 100) {
    power += 0.1;
    Motor::power(0, power);
    Motor::power(3, power);

    auto speeds = Motor::get_instance().get_speed();
    // spdlog::info("{:05.0f} {:05.0f}", speeds.at(0), speeds.at(3));

    // auto start_values = Encoder::read_all();
    // spdlog::info("{}", start_values.at(3));
    // auto start_time = std::chrono::high_resolution_clock::now();
    usleep(1000.0 * (1000.0 / 150.0));
    // auto end_values = Encoder::read_all();
    
    // auto elapsed_time = std::chrono::high_resolution_clock::now() - start_time;
    // long double elapsed_ticks = end_values.at(port) - start_values.at(port);

    // long double elapsed_us = std::chrono::duration_cast<std::chrono::microseconds>(
    //     elapsed_time).count();
    // long double elapsed_s = std::chrono::duration_cast<std::chrono::microseconds>(
    //     elapsed_time).count() / 1000000.0;

    // long double vel_ticks_s = elapsed_ticks / elapsed_s;
    // long double vel_rpm = vel_ticks_s * 60.0L / TICKS_PER_TURN;

    // accumulator = (alpha * vel_rpm) + (1.0 - alpha) * accumulator;

    // spdlog::info("{:05.0f} {:05.0f} {:05.0f} {:05.0f} {:05.0f}", elapsed_us, elapsed_ticks, vel_ticks_s,  vel_rpm, accumulator);
  }


  // spdlog::info(count / (end_time - start_time) * 1000.0);
  // spi.write(Spi::Register::DISPLAY_LINE_1_C0, 1, 66);
  return 0;
}

